//
// Program to generate the pprn format of a 3D table to be protected with
// the controlled tabular adjustment or minimum-distance perturbation method
// Manipulation of TABLE3D is not efficient, but just for testing purposes.
//
// The input format considered is that produced by gen_CSP_instance_3D.
// File is read from standard input

// Usage : gen_pprn_of_3D lb% ub% p% < 3Dtable_in_ampl_format
//
/****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <sys/types.h>
#include <sys/times.h>

// typedefs
typedef struct{
    int m; // number of rows, without marginal
    int n; // number of columns, without marginal
    int l; // number of levels, without marginal
    double *v; /* (m+1)*(n+1)*(l+1) table values (marginals included)
		  from 0..m to 0..n to 0..l */
    int p; // number of primary suppressions
    int *p_m; // rows of the primary suppressions
    int *p_n; // columns of the primary suppressions
    int *p_l; // levels of the primary suppressions
} TABLE3D;


// system dependent macros
#ifdef Linux
#define CLOCKS_PER_SECOND 100  /* for Linux */
#endif
#ifdef SUN
#define CLOCKS_PER_SECOND 100.0  /* for Solaris */
#endif

#define MYALLOC(n,type) (type *)malloc((n)*sizeof(type))
#define MYFREE(p) if (p) {free(p); (p)= NULL;}

// constants
const char TRUE=  1;
const char FALSE= 0;


/* function prototypes */
int allocate_table3D(TABLE3D **ptab, int nrows, int ncols, int nlevs, int p);
void delete_table3D(TABLE3D **ptab);
inline double get_cell3D(TABLE3D *tab, int row, int column, int level);
inline void put_cell3D(TABLE3D *tab, int row, int column, int level, 
		       double value);
int read_table_AMPL_format3D(TABLE3D **tab);
inline int cell_to_arc(TABLE3D *tab, int i, int j, int k);
void transform_to_standard_mmcnf(int nar, int nat, int nnu, int *norig, 
				 int *ndest, double *ub_arc, double *lb_arc,
				 double *cost, double *qcost, double *cmt,
				 double *inj, double *f0shift);
void generate_pprn_format(int nar, int nat, int nnu, int *norig, 
			  int *ndest, double *ub_arc, double *lb_arc,
			  double *cost, double *qcost, double *cmt,
			  double *inj, double f0shift);

int main(int argc, char *argv[])
{  
    TABLE3D *tab= NULL;
    struct tms buffer;
    
    ////
    // check number of command line arguments
    if (argc != 4) {
	fprintf(stderr,"usage: main3D lb%% ub%% p%% < file \n");
	fprintf(stderr,"\t  lb%%:\t %% of lower variation of cell value\n");
	fprintf(stderr,"\t  ub%%:\t %% of upper variation of cell value\n");
	fprintf(stderr,"\t   p%%:\t %% of variation for primaries\n");
	fprintf(stderr,"\t file:\t 3Dtable_in_ampl_format\n");
	delete_table3D(&tab);
	return(-1);
    }

    double lbperc;
    sscanf(argv[1],"%lf",&lbperc);
    if (lbperc <= 0.00000 || lbperc > 100.0) {
	fprintf(stderr,"error: lb%% must be in [0,100] \n");
	return(-1);
    }
    lbperc= 1-lbperc/100;

    double ubperc;
    sscanf(argv[2],"%lf",&ubperc);
    if (ubperc <= 0.00000 || ubperc > 100.0) {
	fprintf(stderr,"error: ub%% must be in [0,100] \n");
	return(-1);
    }
    ubperc= 1+ubperc/100;

    double pperc;
    sscanf(argv[3],"%lf",&pperc);
    if (pperc <= 0.00000 || pperc > 100.0) {
	fprintf(stderr,"error: p%% must be in [0,100] \n");
	return(-1);
    }
    pperc= 1+pperc/100;

    ////
    // create and read table from standard input
    if (read_table_AMPL_format3D(&tab)) {
	fprintf(stderr,"error: reading table\n");
	delete_table3D(&tab);
	return(-1);
    }

    ////
    // create multicommodity problem from 3D table
    // marginal cells are considered fixed (they can not be perturbed)
    // nar*nat arrays: first data of commodity 1, and so on until commodity nat
    // nnu*nat arrays: first data of commodity 1, and so on until commodity nat

    int nnu= (tab->m)+(tab->n);
    int nar= (tab->m)*(tab->n);
    int nat= (tab->l);
    int ncb= 0;
    int nlb= 0;
    int *norig= new int[nar];
    int *ndest= new int[nar];
    double *ub_arc= new double[nar*nat];
    double *lb_arc= new double[nar*nat];
    double *cost= new double[nar*nat];
    double *qcost= new double[nar*nat];
    double *cmt= new double[nar];
    double *inj= new double[nnu*nat];
    double *flow= new double[nar*nat];

    // network topology (origin and dest. nodes for arc) and mutual capacity
    int i,j,k,ct;
    ct= 0;
    for(i=0;i<tab->m;i++) 
	for(j=0;j<tab->n;j++) {
	    norig[ct]= i;
	    ndest[ct]= tab->m+j;
	    cmt[ct]= get_cell3D(tab,i,j,tab->l);
	    ct++;
	}
	 
    // arcs data: bounds and costs
    double cellvalue, f0= 0;
    ct= 0;
    for(k=0;k<tab->l;k++) 
	for(i=0;i<tab->m;i++) 
	    for(j=0;j<tab->n;j++) {
		cellvalue= get_cell3D(tab,i,j,k);
		lb_arc[ct]= lbperc*cellvalue;
		if (cellvalue<1.0e-16) // check if is a zero cell
		    ub_arc[ct]= 0.0;
		else
		    ub_arc[ct]= ubperc*cellvalue;
		f0 += cellvalue*cellvalue; // fixed term for objective function
		qcost[ct]= 2.0; // 1/2 x'Qx is considered as quadratic term 
		cost[ct]= -2*cellvalue;
		ct++;
	    }
    // nodes data: injections
    ct= 0;
    for(k=0;k<tab->l;k++) {
	for(i=0;i<tab->m;i++) {
	    inj[ct]= get_cell3D(tab,i,tab->n,k); 
	    ct++;
	}	
 	for(j=0;j<tab->n;j++) {
	    inj[ct]= -get_cell3D(tab,tab->m,j,k);
	    ct++;
	}
    }
    
    // adjust primary cells bounds
    for(i=0;i<tab->p;i++) {
	cellvalue= get_cell3D(tab,tab->p_m[i],tab->p_n[i],tab->p_l[i]);
	ct= cell_to_arc(tab,tab->p_m[i],tab->p_n[i],tab->p_l[i]);
	lb_arc[ct]= pperc*cellvalue;
	ub_arc[ct]= pperc*cellvalue;
    }

    // transform multicommodity problem to a standard form with zero
    // lower bounds
    double f0shift;
    transform_to_standard_mmcnf(nar,nat,nnu,norig,ndest,ub_arc,lb_arc,
				cost,qcost,cmt,inj,&f0shift);
    // write to standard output
    generate_pprn_format(nar,nat,nnu,norig,ndest,ub_arc,lb_arc,
			 cost,qcost,cmt,inj,f0shift);
 
    
    delete_table3D(&tab);
    return(0);

}

int allocate_table3D(TABLE3D **ptab, int nrows, int ncols, int nlevs, int p)
{
    (*ptab)= new TABLE3D;
    (*ptab)->m= nrows;
    (*ptab)->n= ncols;
    (*ptab)->l= nlevs;
    (*ptab)->p= p;
    (*ptab)->v= new double[(nrows+1)*(ncols+1)*(nlevs+1)];
    (*ptab)->p_m= new int[p];
    (*ptab)->p_n= new int[p];
    (*ptab)->p_l= new int[p];
    return(0);
}

void delete_table3D(TABLE3D **ptab) 
{
    if (*ptab) {
	delete [] (*ptab)->v;
	delete [] (*ptab)->p_m;
	delete [] (*ptab)->p_n;
	delete [] (*ptab)->p_l;
	delete [] (*ptab);
    }
}

inline double get_cell3D(TABLE3D *tab, int row, int column, int level)
{
    return(tab->v[level*(tab->n+1)*(tab->m+1)+row*(tab->n+1)+column]);
}

inline void put_cell3D(TABLE3D *tab, int row, int column, int level, 
		       double value)
{
    tab->v[level*(tab->n+1)*(tab->m+1)+row*(tab->n+1)+column]= value;
}

/* create and read table in AMPL format from filename */
// returns 0 if everything goes fine
// returns -1 if problems creating table3D object
// returns -2 if file not found
int read_table_AMPL_format3D(TABLE3D **ptab)
{
    int nrows, ncols, nlevs, p;
    const int MJUNK= 1000;
    char junk[MJUNK];
    int i,j,k;
    double vcell;
    int icell, jcell, kcell;

    // read table dimensions
  
    gets(junk);
    gets(junk);
    gets(junk);
    gets(junk);
    scanf("%*s %*s %*s %d ",&nrows);gets(junk);
    gets(junk);
    gets(junk);
    scanf("%*s %*s %*s %d ",&ncols);gets(junk);
    gets(junk);
    gets(junk);
    scanf("%*s %*s %*s %d ",&nlevs);gets(junk);
    gets(junk);
    gets(junk);
    scanf("%*s %*s %*s %d ",&p);gets(junk);

    // add marginals
    if (allocate_table3D(ptab,nrows,ncols,nlevs,p)) return(-1);
    TABLE3D *tab= *ptab;

    //read table values (entries and primary cells)
    gets(junk);
    gets(junk);
    gets(junk);
    int pi,pj,pl;
    double plpl,pupl;
    for(i= 0;i<p;i++) {
	scanf("%*d %d %d %d %lf %lf\n",&pi,&pj,&pl,&plpl,&pupl);
	tab->p_m[i]= pi-1;
	tab->p_n[i]= pj-1;
	tab->p_l[i]= pl-1;
    }
    gets(junk);
    gets(junk);
    gets(junk);
    gets(junk);
    for(i= 0;i<tab->m+1;i++) 
	for (j=0;j<tab->n+1;j++) 
	    for (k= 0;k<tab->l+1;k++) {
		scanf("%d %d %d %lf\n",&icell,&jcell,&kcell,&vcell);
		icell--;
		jcell--;
		kcell--;
		put_cell3D(tab,icell,jcell,kcell,vcell);
	    }

// //writes table for testing purposes
//     for(i= 0;i<p;i++) {
// 	printf("%d \t %d \t %d \t %d\n",i+1,tab->p_m[i]+1,tab->p_n[i]+1,
// 	       tab->p_l[i]+1);
//     }
//     printf("\n\n");
//     for(i= 0;i<tab->m+1;i++) 
// 	for (j=0;j<tab->n+1;j++) 
// 	    for (k= 0;k<tab->l+1;k++) {
// 		printf("%d \t %d \t %d \t %lf\n",i+1,j+1,k+1,
// 		       get_cell3D(tab,i,j,k));
// 	    }
    
    
    return(0);
}

inline int cell_to_arc(TABLE3D *tab, int i, int j, int k)
{
    return(k*(tab->m)*(tab->n)+i*(tab->n)+j);
}



void transform_to_standard_mmcnf(int nar, int nat, int nnu, int *norig, 
				 int *ndest, double *ub_arc, double *lb_arc,
				 double *cost, double *qcost, double *cmt,
				 double *inj, double *f0shift)
// transform the problem to one with 0 lower bounds through the change of 
// variables x'=x-lb. Array lb_arc is not modified.
{
    int i,k;
    
    //upper bounds (lower not changed)
    for(i=0;i<nar*nat;i++) ub_arc[i] -= lb_arc[i];
    
    // mutual capacity
    int ct= 0;
    for(k=0;k<nat;k++) 
	for(i=0;i<nar;i++) {
	    cmt[i] -= lb_arc[ct];
	    ct++;
	}
    
    // injections
    ct= 0;
    int ctnnu= 0;
    for(k=0;k<nat;k++) {
	for(i=0;i<nar;i++) {
	    inj[ctnnu+norig[i]] -= lb_arc[ct]; //arc leaving node
	    inj[ctnnu+ndest[i]] += lb_arc[ct];// arc entering node
	    ct++;
	}
	ctnnu += nnu;
    }

    //costs
    *f0shift= 0;
    for(i=0;i<nar*nat;i++) {
	// in this order: first f0shift, later update linear cost
	(*f0shift) += (0.5*qcost[i]*lb_arc[i]+cost[i])*lb_arc[i];
	cost[i] += qcost[i]*lb_arc[i];
    }

    return;
}


void generate_pprn_format(int nar, int nat, int nnu, int *norig, 
			  int *ndest, double *ub_arc, double *lb_arc,
			  double *cost, double *qcost, double *cmt,
			  double *inj, double f0shift)    
{
    int k,i,ct;
    cout << nnu << " "<< nar << " "<<  nat << " "<< 0 << " "<< 0 << " "<< " "<< nar*nat << "\n" ;
    cout << "linear costs per commodity \n";
    ct=0;
    for(k=0;k<nat;k++) {
	for(i=0;i<nar;i++) {
	    cout << cost[ct] << " ";
	    ct++;
	}
	cout << "\n";
    }
    cout << "quadratic costs per commodity \n";
    ct=0;
    for(k=0;k<nat;k++) {
	for(i=0;i<nar;i++) {
	    cout << qcost[ct]/2.0 << " "; // divided by 2 because pprn format
	                                  // considers x'Qx objectives 
	    ct++;
	}
	cout << "\n";
    }
    cout << "arc capacities per commodity \n";
    ct=0;
    for(k=0;k<nat;k++) {
	for(i=0;i<nar;i++) {
	    // change zero capacities by small value (required by QIPM)
	    if (ub_arc[ct]<1.0e-16)
		cout << 1.0e-16 << " ";
	    else
		cout << ub_arc[ct] << " "; 
	    ct++;
	}
	cout << "\n";
    }
    cout << "node supply and demands per commodity \n";
    ct=0;
    for(k=0;k<nat;k++) {
	for(i=0;i<nnu;i++) {
	    cout << inj[ct] << " "; 
	    ct++;
	}
	cout << "\n";
    }
    cout << "mutual arc capacity for all the commodities \n";
    for(i=0;i<nar;i++)
	cout << cmt[i]<< " ";
    cout << "\n";
    cout << "network topology: origin and destination per arc \n";
    for(i=0;i<nar;i++)
	cout << norig[i]+1 <<"\t"<< ndest[i]+1 << "\n";
    cout << "\n";
    
}




